/** @file wssecurity.c
 * \brief Web Services Security implementation for Cohesion web services. 
 *        See http://www.oasis-open.org/committees/wss/.
 *
 * This is a simplified API to xmlsec (http://www.aleksey.com/xmlsec) 
 * library.  Two public methods are provided:
 *
 * - wssecurity_sign_mem() which sign a SOAP message, adding required
 *   signature nodes to the xml document;
 * - wssecurity_verify_mem() which verify a SOAP message.
 *
 * \warning <tt>[2006-12-12]</tt> This IS NOT a full implementation of
 * Oasis Web Services Security specifications. It is known to work
 * only with a particular application (Cohesion, see
 * http://www.rete.marche.it/sut/sut.asp?pag=15_1) based on MS .NET
 * web services.
 *
 * @author Davide Principi <d.principi(AT)provincia.ps.it>
 * @author Based on "verify3.c" (C) 2002-2003 Aleksey Sanin <aleksey(AT)aleksey.com>
 * @version \verbatim $Id: wssecurity.c 1692 2006-12-14 07:31:49Z davide $ \endverbatim
 */

/** @mainpage Licensing
 * Copyright (C) 2006 <b>Amministrazione Provinciale di Pesaro e Urbino</b> - Italy  
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/** @page install Installation
 * \verbinclude INSTALL.txt
 */

/** @page gpl General Public License
 * \verbinclude GPL.txt
 */


#include <stdlib.h>
#include <string.h>
#include <assert.h>
 
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
 
#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/xmldsig.h>
#include <xmlsec/xmlenc.h>
#include <xmlsec/crypto.h>
#include <xmlsec/bn.h>
#include <xmlsec/base64.h>

#ifndef __WSSECURITY_H__
#include "wssecurity.h"


#define WSSECURITY_ERRMSG_PREFIX "wssecurity:"
#ifndef WSSECURITY_BUILD_DATE
#define WSSECURITY_BUILD_DATE "not set"
#endif


const char *const WSSECURITY_Rev = "$Rev: 1692 $";
const char *const WSSECURITY_Date = "$Date: 2006-12-14 08:31:49 +0100 (Thu, 14 Dec 2006) $";
const char *const WSSECURITY_Build_Date = WSSECURITY_BUILD_DATE;
const char *const WSSECURITY_XMLSEC_Version = XMLSEC_VERSION;

#ifdef LIBXML_DOTTED_VERSION
const char *const WSSECURITY_LIBXML_Version = LIBXML_DOTTED_VERSION;
#else
const char *const WSSECURITY_LIBXML_Version = "unknown";
#endif

#ifdef OPENSSL_VERSION_TEXT
const char *const WSSECURITY_OPENSSL_Version = OPENSSL_VERSION_TEXT;
#else 
const char *const WSSECURITY_OPENSSL_Version = "dynamic";
#endif

const char *const WSSECURITY_SECEXT_NS = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";

 
static int _wss_lib_init() ;
static int _wss_lib_cleanup() ;
static int _wss_xmlSecAppAddIDAttr(xmlNodePtr, const xmlChar* , const xmlChar* ) ;
static xmlNodePtr _wss_find_bst(xmlNodePtr doc);
static xmlSecKeysMngrPtr _wss_load_trusted_certs(char** files);

/**
 * Sign message in memory.
 * @param templateMsg The SOAP message to be signed.
 * @param securityTokenID The ID of the SecurityToken element.
 * @param idList The IDs of the elements to be signed. Last element MUST be NULL.
 * @param keyFile Path to the private pem certificate.
 * @param password Private certificate password.
 * @param certFile Path to the pem certificate.
 * @return Signed message, or NULL, if failed.  Memory must be <CODE>free</CODE>d.
 * @see wssecurity_verify_mem()
 */
char* 
wssecurity_sign_mem(const char* templateMsg,
		    const char *securityTokenID, 
		    char *idList[],
		    const char* keyFile, 
		    const char* password,
		    const char* certFile) { 
  
  xmlDocPtr doc = NULL;

  xmlNodePtr node_signature = NULL; /* Signature */
  xmlNodePtr node_security = NULL; /* Security */
  xmlNodePtr node_keyinfo = NULL; /* KeyInfo */
  xmlNodePtr node_stkr = NULL; /* Reference node inside SecurityTokenReference */
  xmlNodePtr node_bst = NULL; /* BinarySecurityToken */
  xmlNodePtr node_str = NULL; /* SecurityTokenReference */

  xmlNsPtr ns_wsse = NULL;

  xmlSecDSigCtxPtr dsigCtx = NULL;
  char *result = NULL;

  xmlSecKeyInfoCtxPtr keyInfoCtx_bst = NULL; /* binary security token context */

  int retval = -1;
  int i = 0;
    
  retval = _wss_lib_init();
  if (retval < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "FAILED 1\n");
    return(NULL);
  } 

  keyInfoCtx_bst = xmlSecKeyInfoCtxCreate(NULL);
  if(keyInfoCtx_bst == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "FAILED 2\n");
    return(NULL);
  }
  xmlSecKeyInfoCtxInitialize(keyInfoCtx_bst, NULL);
  keyInfoCtx_bst->base64LineSize = 0;

  /* load template */
  doc = xmlParseDoc((xmlChar *) templateMsg);
  if ((doc == NULL) || (xmlDocGetRootElement(doc) == NULL)){
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: unable to parse file \"%s\"\n", templateMsg);
    goto done;    
  }

  if(1) {
    retval = _wss_xmlSecAppAddIDAttr(xmlDocGetRootElement(doc), (xmlChar*) "Id", NULL);
    if(retval < 0) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to add new ID attriubte\n");
      goto done;
    }
  }

  /* create <Signature></Signature> template node_signature */
  node_signature = xmlSecTmplSignatureCreate(doc, xmlSecTransformExclC14NId, xmlSecTransformRsaSha1Id, NULL);

  /* find <wsse:Security> */
  node_security = xmlSecFindNode(xmlDocGetRootElement(doc), (xmlChar*) "Security", (xmlChar*) WSSECURITY_SECEXT_NS); 
  if(node_security == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: Security node not found\n");
    goto done;
  }

  /* find <wsse:BinarySecurityToken> */
  node_bst = xmlSecFindChild(node_security, (xmlChar*) "BinarySecurityToken", (xmlChar*) WSSECURITY_SECEXT_NS);
  if(node_bst == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: BinarySecurityToken node not found\n");
    goto done;
  }

  /* append node to <Security> */
  node_signature = xmlSecAddChildNode(node_security, node_signature);

  for(i = 0; idList != NULL && idList[i] != NULL; i++) {
    char* id_reference = NULL;
    char id_buf[256];
    xmlNodePtr node_reference = NULL;

    id_buf[0] = '#';
    id_buf[1] = '\0';

    id_reference = idList[i];

    node_reference = xmlSecTmplSignatureAddReference(node_signature, xmlSecTransformSha1Id, NULL, NULL, NULL);

    if(node_reference == NULL) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to create reference node %s\n", id_reference);
      goto done;
    }

    xmlSetProp(node_reference, (xmlChar*) "URI", (xmlChar*) strncat(id_buf, id_reference, 255));

    node_reference = xmlSecTmplReferenceAddTransform(node_reference, xmlSecTransformExclC14NId);
  }

  /* add KeyInfo node */
  node_keyinfo = xmlSecTmplSignatureEnsureKeyInfo(node_signature, NULL);
  if(node_keyinfo == NULL) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to add keyinfo node\n");
      goto done;
  }

  ns_wsse = xmlSearchNsByHref(doc, node_keyinfo, (xmlChar*) WSSECURITY_SECEXT_NS);
  if(ns_wsse == NULL) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: oasis wss-wssecurity-secext namespace not found\n");
      goto done;
  }

  node_str = xmlNewChild(node_keyinfo, ns_wsse, (xmlChar*) "SecurityTokenReference", NULL);

  /* Add reference to BinarySecurityToken node */
  node_stkr = xmlNewChild(node_str, ns_wsse, (xmlChar*) "Reference", NULL);
  xmlSetProp(node_stkr, (xmlChar*) "URI", (xmlChar*) securityTokenID);
  xmlSetProp(node_stkr, (xmlChar*) "ValueType", (xmlChar*) "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509v3");

  /* create signature context, we don't need keys manager in this example */
  dsigCtx = xmlSecDSigCtxCreate(NULL);
  if(dsigCtx == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to create signature context\n");
    goto done;
  }
        
  /* load private key */
  dsigCtx->signKey = xmlSecCryptoAppKeyLoad(keyFile, xmlSecKeyDataFormatPem, password, NULL, NULL);
  if(dsigCtx->signKey == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to load private pem key from \"%s\"\n", keyFile);
    goto done;
  }
        
  if(certFile != NULL) {
    if(xmlSecCryptoAppKeyCertLoad(dsigCtx->signKey, certFile, xmlSecKeyDataFormatPem) < 0) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to load pem certificate \"%s\"\n", certFile);
      goto done;
    }
  }
        
/*   /\* set key name to the file name, this is just an example! *\/ */
/*   if(xmlSecKeySetName(dsigCtx->signKey, keyFile) < 0) { */
/*     fprintf(stderr,"Error: failed to set key name for key from \"%s\"\n", keyFile); */
/*     goto done; */
/*   } */


  xmlSecKeyDataXmlWrite(xmlSecKeyDataX509Id, dsigCtx->signKey, node_keyinfo, keyInfoCtx_bst);

  /* Move X509Certificate child nodes to BinarySecurityToken */
  {

    xmlNodePtr node_x509cert = NULL;
    xmlNodePtr cur = NULL;  

    node_x509cert = xmlSecFindChild(node_keyinfo, (xmlChar*) "X509Certificate", xmlSecDSigNs); 

    if(node_x509cert == NULL) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "X509Certificate node not found.\n");    
      goto done;
    }

    for(cur = node_x509cert->children;
	cur != NULL;
	cur = xmlSecGetNextElementNode(cur)) {
      xmlUnlinkNode(cur);
      xmlAddChild(node_bst, cur);
    }

    xmlUnlinkNode(node_x509cert);
    xmlFreeNode(node_x509cert);
    node_x509cert = NULL;
  }

                
  /* sign the template */
  if(xmlSecDSigCtxSign(dsigCtx, node_signature) < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "FAILED 3\n");
    goto done;
  }

  if(dsigCtx->status != xmlSecDSigStatusSucceeded) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "FAILED 4\n");
    goto done;
  }

  {
    /* print signed document to char* for return to caller */        
    xmlChar *output = NULL;
    int output_sz = 0;

    xmlDocDumpMemory(doc, &output, &output_sz);
    result = strndup(output, output_sz);
    xmlFree(output);
  }

 done:    

  /* cleanup */

  if(keyInfoCtx_bst != NULL) {
    /* Called by destroy: xmlSecKeyInfoCtxFinalize(keyInfoCtx_bst); */
    xmlSecKeyInfoCtxDestroy(keyInfoCtx_bst);
  }
  
  if(dsigCtx != NULL) {
    xmlSecDSigCtxDestroy(dsigCtx);
  }
  
  if(doc != NULL) {
    xmlFreeDoc(doc);
  }
  
  _wss_lib_cleanup() ;
 
  return((char *) result);
}


/**
 * Verify message signature.
 * @param xmlMessage The message to be verified.
 * @param keyFileList Certificate list used to check the signature. Last element MUST be NULL.
 * @return zero on success, non-zero otherwise. 
 * @see wssecurity_sign_mem()
 */
int 
wssecurity_verify_mem(const char* xmlMessage, 
		      char** keyFileList) {

  xmlDocPtr doc = NULL;
  xmlSecDSigCtxPtr dsigCtx = NULL;
  xmlSecKeysMngrPtr mngr_def = NULL;

  int retval = -1;
  int signstatus = -1;

  xmlNodePtr node_signature = NULL; /* Signature */
  xmlNodePtr node_security = NULL; /* Security */
  xmlNodePtr node_keyinfo = NULL; /* KeyInfo */
  xmlNodePtr node_stkr = NULL; /* Reference node inside SecurityTokenReference */
  xmlNodePtr node_bst = NULL; /* BinarySecurityToken */
  xmlNodePtr node_str = NULL; /* SecurityTokenReference */

  assert(xmlMessage);

  retval = _wss_lib_init();
  if(retval < 0) {
    return -2;
  }

  mngr_def = _wss_load_trusted_certs(keyFileList);
  if(mngr_def == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Key manager creation failed.\n");
    goto done;
  }


  /* parse document */
  doc = xmlParseDoc((xmlChar *) xmlMessage) ;
 
  if ((doc == NULL) || (xmlDocGetRootElement(doc) == NULL)){
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: unable to parse file \"%s\"\n", xmlMessage);
    goto done;
  }

  if(1) {
    /* add Id attribute as xml ID */
    retval = _wss_xmlSecAppAddIDAttr(xmlDocGetRootElement(doc), (xmlChar*) "Id", NULL);
    if(retval < 0) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to add new ID attriubte\n");
      goto done;
    }
  }
    
  /* find start node */
  node_signature = xmlSecFindNode(xmlDocGetRootElement(doc), xmlSecNodeSignature, xmlSecDSigNs);
  if(node_signature == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: start node not found.\n");
    goto done;
  }

  /* find BinarySecurityToken node */
  node_bst = _wss_find_bst(xmlDocGetRootElement(doc));
  if(node_bst == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "BinarySecurityToken node not found\n");
    goto done;
  }
  
  /* find KeyInfo node */
  node_keyinfo = xmlSecFindChild(node_signature, (xmlChar*) "KeyInfo", xmlSecDSigNs);
  if(node_keyinfo == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: KeyInfo node not found\n");
    goto done;
  }


  /* Copy BinarySecurityToken contents to X509Certificate */
  {
    xmlNodePtr node_x509cert = NULL;
    xmlNodePtr node_x509data = NULL;
    xmlNodePtr cur = NULL;  

    node_x509data = xmlNewChild(node_keyinfo, NULL, (xmlChar*) "X509Data", NULL);
    node_x509cert = xmlNewChild(node_x509data, NULL, (xmlChar*) "X509Certificate", NULL);

    for(cur = node_bst->children;
	cur != NULL;
	cur = xmlSecGetNextElementNode(cur)) {
      xmlAddChild(node_x509cert, xmlCopyNode(cur, 0));
    }
  }


  /* create signature context */
  dsigCtx = xmlSecDSigCtxCreate(mngr_def);
  if(dsigCtx == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to create signature context\n");
    goto done;
  }

  /* Verify signature */
  if(xmlSecDSigCtxVerify(dsigCtx, node_signature) < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: signature verify\n");
    goto done;
  }
        
  /* prepare return code */
  if(dsigCtx->status == xmlSecDSigStatusSucceeded) {
    signstatus = 0;
  } else {
    signstatus = 1;
  }    

 done:    
  /* cleanup */

  if(mngr_def != NULL) {
    xmlSecKeysMngrDestroy(mngr_def);
  }

  if(dsigCtx != NULL) {
    xmlSecDSigCtxDestroy(dsigCtx);
  }
    
  if(doc != NULL) {
    xmlFreeDoc(doc); 
  }

  _wss_lib_cleanup() ; 
  return(signstatus);
}

/**
 * _wss_find_bst: find BinarySecurityToken node 
 * TODO: use XPATH to
 * find the node, taking info from wsse:Reference node in
 * SecurityTokenReference in KeyInfo node
 */
static xmlNodePtr _wss_find_bst(xmlNodePtr startNode) {

  xmlNodePtr node_bst = NULL;

  node_bst = xmlSecFindNode(startNode, (xmlChar*) "BinarySecurityToken", (xmlChar *) WSSECURITY_SECEXT_NS);
  return(node_bst);
}

/**
 * Creates simple keys manager and load trusted certificates from PEM #files.
 * The caller is responsible for destroing returned keys manager using
 * xmlSecKeysMngrDestroy.
 *
 * @param files An array of file names. Last element MUST be NULL.
 * @return the pointer to newly created keys manager or NULL if an error
 * occurs.
 */
static xmlSecKeysMngrPtr 
_wss_load_trusted_certs(char **files) {
  xmlSecKeysMngrPtr mngr;
  int i = 0;
        
  assert(files);
    
  /* create and initialize keys manager, we use a simple list based
   * keys manager, implement your own xmlSecKeysStore klass if you need
   * something more sophisticated 
   */
  mngr = xmlSecKeysMngrCreate();
  if(mngr == NULL) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: failed to create keys manager.\n");
    return(NULL);
  }
  if(xmlSecCryptoAppDefaultKeysMngrInit(mngr) < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: failed to initialize keys manager.\n");
    xmlSecKeysMngrDestroy(mngr);
    return(NULL);
  }    
    
  for(i = 0; files[i] != NULL; i++) {

    /* load trusted cert */
    if(xmlSecCryptoAppKeysMngrCertLoad(mngr, files[i], xmlSecKeyDataFormatPem, xmlSecKeyDataTypeTrusted) < 0) {
      fprintf(stderr, WSSECURITY_ERRMSG_PREFIX "Error: failed to load pem certificate from \"%s\"\n", files[i]);
      xmlSecKeysMngrDestroy(mngr);
      return(NULL);
    }
  }

  return(mngr);
}


/**
 * Set recursively Id attribute as ID.
 */
static int  
_wss_xmlSecAppAddIDAttr(xmlNodePtr node, const xmlChar* attrName, const xmlChar* nsHref) {
  xmlAttrPtr attr, tmpAttr;
  xmlNodePtr cur;
  xmlChar* id;
    
  if((node == NULL) || (attrName == NULL)) {
    return(-1);
  }
    
  /* process children first because it does not matter much but does simplify code */
  cur = xmlSecGetNextElementNode(node->children);
  while(cur != NULL) {
    if(_wss_xmlSecAppAddIDAttr(cur, attrName, nsHref) < 0) {
      return(-1);
    }
    cur = xmlSecGetNextElementNode(cur->next);
    /*if(cur != NULL && attrName != NULL && cur->name != NULL) {
      fprintf(stderr, "Adding %s attribute to node %s\r\n", attrName, cur->name);
      }*/
  }
    
  
  /* if nsHref is set then it also should match */    
  if((nsHref != NULL) && (node->ns != NULL) && (!xmlStrEqual(nsHref, node->ns->href))) {
    return(0);
  }
    
  /* the attribute with name equal to attrName should exist */
  for(attr = node->properties; attr != NULL; attr = attr->next) {
    if(xmlStrEqual(attr->name, attrName)) {
      break;
    }
  }
  if(attr == NULL) {
    return(0);
  }
    
  /* and this attr should have a value */
  id = xmlNodeListGetString(node->doc, attr->children, 1);
  if(id == NULL) {
    return(0);
  }
    
  /* check that we don't have same ID already */
  tmpAttr = xmlGetID(node->doc, id);
  if(tmpAttr == NULL) {
    xmlAddID(NULL, node->doc, id, attr);
  } else if(tmpAttr != attr) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: duplicate ID attribute \"%s\"\n", id);
    xmlFree(id);
    return(-1);
  }
  xmlFree(id);
  return(0);
}

 
static int 
_wss_lib_init()
{
  /* Init libxml and libxslt libraries */
  xmlInitParser();
  LIBXML_TEST_VERSION
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
  xmlSubstituteEntitiesDefault(1);
#ifndef XMLSEC_NO_XSLT
  xmlIndentTreeOutput = 1;
#endif /* XMLSEC_NO_XSLT */
            
  /* Init xmlsec library */
  if(xmlSecInit() < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: xmlsec initialization failed.\n");
    return(-1);
  }
 
  /* Check loaded library version */
  if(xmlSecCheckVersion() != 1) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: loaded xmlsec library version is not compatible.\n");
    return(-1);
  }
 
  /* Load default crypto engine if we are supporting dynamic
   * loading for xmlsec-crypto libraries. Use the crypto library
   * name ("openssl", "nss", etc.) to load corresponding
   * xmlsec-crypto library.
   */
#ifdef XMLSEC_CRYPTO_DYNAMIC_LOADING
  if(xmlSecCryptoDLLoadLibrary(BAD_CAST XMLSEC_CRYPTO) < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  
	    "Error: unable to load default xmlsec-crypto library. Make sure\n"
            "that you have it installed and check shared libraries path\n"
            "(LD_LIBRARY_PATH) envornment variable.\n");
    return(-1);    
  }
#endif /* XMLSEC_CRYPTO_DYNAMIC_LOADING */
 
  /* Init crypto library */
  if(xmlSecCryptoAppInit(NULL) < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: crypto initialization failed.\n");
    return(-1);
  }
 
  /* Init xmlsec-crypto library */
  if(xmlSecCryptoInit() < 0) {
    fprintf(stderr, WSSECURITY_ERRMSG_PREFIX  "Error: xmlsec-crypto initialization failed.\n");
    return(-1);
  }

  return(0);
}
 
static int 
_wss_lib_cleanup()
{
  /* Shutdown xmlsec-crypto library */
  xmlSecCryptoShutdown();
  
  /* Shutdown crypto library */
  xmlSecCryptoAppShutdown();
  
  /* Shutdown xmlsec library */
  xmlSecShutdown();
  
  /* Shutdown libxslt/libxml */
#ifndef XMLSEC_NO_XSLT
  xsltCleanupGlobals();            
#endif /* XMLSEC_NO_XSLT */
  xmlCleanupParser();

  return(0);
}

#endif /* __WSSECURITY_H__ */
